import typer
import rich
from rich.prompt import Confirm
from rich.table import Table
from sqlalchemy.exc import IntegrityError

import port_forward_manager.cli_schema
from .cli_schema import prepare_schema_table
from . import database, models

app = typer.Typer(no_args_is_help=True)
db = database.SessionLocal()


def prepare_group_table(show_alias: bool = False, show_link: bool = False, show_schema: bool = True, title: str = 'Schemas'):

    table = Table(show_edge=False)

    table.add_column("#", justify="right", style="green", width=3)
    table.add_column("Name", justify="left", style="yellow", width=30)
    table.add_column("Label", justify="right", style="yellow", width=30)

    return table


@app.command()
def index(name=typer.Argument(None, help='Criteria to search')):
    """
    List groups
    """
    table = prepare_group_table()
    for group in models.Group.index(name):
        row = [
            str(group.id),
            group.name,
            group.label,
        ]
        table.add_row(*row)

    rich.print(table)


@app.command()
def schemas(name=typer.Argument(None, help='Criteria to search')):
    """
    List groups
    """
    table = prepare_schema_table()
    for group in models.Group.index(name):
        for schema in group.schemas:
            row = [
                str(schema.id),
                group.name,
                schema.name,
                schema.label
            ]
            table.add_row(*row)

    rich.print(table)


@app.command()
def create(name=typer.Argument(None, help='Group unique name'),
           label=typer.Argument(None, help='Group label')):
    """
    Create a group
    """
    group = models.Group(name=name, label=label)
    db.add(group)
    try:
        db.commit()
    except IntegrityError as e:
        print(f"Name MUST be unique {e.args}")
    except Exception as ex:
        template = "An exception of type {0} occurred. Arguments:\n{1!r}"
        message = template.format(type(ex).__name__, ex.args)
        print(message)


@app.command()
def update(group_id=typer.Argument(None, help="Group ID"),
           name=typer.Option(None, help='Group unique name'),
           label=typer.Option(None, help='Group label')):
    """
    Update a group
    """
    group = models.Group.find_by_id(group_id)
    if not group:
        print("Session ID not found")
        exit(401)

    if name:
        group.name = name

    if label:
        group.label = label

    try:
        models.db_session.commit()
    except IntegrityError as e:
        print(f"Name MUST be unique {e.args}")
    except Exception as ex:
        template = "An exception of type {0} occurred. Arguments:\n{1!r}"
        message = template.format(type(ex).__name__, ex.args)
        print(message)


@app.command()
def delete(group_id: int = typer.Argument(None, help='Group name'),
           force: bool = typer.Option(False, help="Force delete")):
    """Delete session"""

    group = models.Group.find_by_id(group_id)
    if group is None:
        print(f"Group '{group_id}' not found")
        exit()

    if force or Confirm.ask(f"Are you sure you want to delete '{group.label}'?"):
        models.Group.delete(group)



from typing import List
from sqlalchemy import Boolean, Column, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

from .database import BaseModelDB, engine, SessionLocal

db_session = SessionLocal()


class Group(BaseModelDB):
    name = Column(String, unique=True, nullable=False)
    label = Column(String)
    schemas = relationship('Schema', back_populates='group', cascade="all,delete")

    @staticmethod
    def delete(record: 'Group'):
        db_session.delete(record)
        db_session.commit()

    @staticmethod
    def index(name=None):
        cursor = db_session.query(Group)
        if name:
            search = "%{}%".format(name)
            cursor = cursor.filter(Group.name.like(search))

        return cursor.all()

    @staticmethod
    def find_by(**kwargs):
        cursor = db_session.query(Group)
        return cursor.filter_by(**kwargs).first()

    @staticmethod
    def find_by_id(group_id: int):
        cursor = db_session.query(Group)
        return cursor.filter(Group.id == group_id).first()

    @staticmethod
    def find_by_name(name):
        cursor = db_session.query(Group)
        return cursor.filter(Group.name == name).first()

    @staticmethod
    def add(group: 'Group'):
        db_session.add(group)
        db_session.commit()

    @classmethod
    def query(cls):
        return db_session.query(cls)

    @staticmethod
    def get_state():
        state = []
        for group in Group.index():
            state.append(group.as_dict())

        return state


class Schema(BaseModelDB):
    name = Column(String, unique=True, nullable=False)
    label = Column(String)
    active = Column(Boolean, index=True)
    group_id = Column(Integer, ForeignKey('group.id'))
    group = relationship('Group', back_populates='schemas')
    sessions = relationship('Session', cascade="all,delete", back_populates='schema')
    ssh_groups = relationship('SSHGroup', cascade="all,delete", back_populates='schema')

    @staticmethod
    def delete(schema: 'Schema'):
        db_session.delete(schema)
        db_session.commit()

    @staticmethod
    def index(name=None, group_id=None):
        cursor = db_session.query(Schema)
        if name:
            search = "%{}%".format(name)
            cursor = cursor.filter(Schema.name.like(search))

        if group_id:
            cursor = cursor.filter_by(group_id=group_id)

        return cursor.all()

    @staticmethod
    def find_by_name(name):
        cursor = db_session.query(Schema)
        return cursor.filter(Schema.name == name).first()

    @staticmethod
    def find_by_id(schema_id: int):
        cursor = db_session.query(Schema)
        return cursor.filter(Schema.id == schema_id).first()

    def get_session(self, session_type: str, hostname: str, remote_port: int):
        for session in self.sessions:
            # print(session.dict())
            # print(hostname, remote_port)

            if session.hostname == hostname and session.remote_port == remote_port and session_type == session.type:
                # print(session.as_dict())
                return session

        # raise Exception(f"Session {hostname} and {remote_port} not found on {self.id}")

    def get_ssh_group(self, group_name: str):
        for group in self.ssh_groups:
            # print(session.dict())
            # print(hostname, remote_port)

            if group.group_name == group_name:
                # print(session.as_dict())
                return group

    @staticmethod
    def get_state():
        state = []
        for schema in Schema.index():
            schema_state = schema.as_dict()

            state.append(schema_state)

        return state


class SSHGroup(BaseModelDB):
    label = Column(String)
    schema_id = Column(Integer, ForeignKey('schema.id'))
    group_name = Column(String)
    schema = relationship('Schema', back_populates='ssh_groups')

    @staticmethod
    def index(group_name=None) -> List['SSHGroup']:
        cursor = db_session.query(SSHGroup)
        if group_name:
            search = "%{}%".format(group_name)
            cursor = cursor.filter(SSHGroup.group_name.like(search))

        return cursor.all()

    @staticmethod
    def find_by_id(group_id: int):
        cursor = db_session.query(SSHGroup)
        return cursor.filter(SSHGroup.id == group_id).first()

    @staticmethod
    def delete(group: 'SSHGroup'):
        db_session.delete(group)
        db_session.commit()

    @staticmethod
    def get_state():
        state = []
        for group in SSHGroup.index():
            group_state = group.as_dict()

            state.append(group_state)

        return state


class Session(BaseModelDB):
    label = Column(String)
    schema_id = Column(Integer, ForeignKey('schema.id'))
    active = Column(Boolean)
    hostname = Column(String)
    type = Column(String, default='local')
    remote_address = Column(String, default='127.0.0.1')
    remote_port = Column(Integer)
    local_address = Column(String, default='127.0.0.1')
    local_port = Column(Integer)
    local_port_dynamic = Column(Boolean, default=False)
    url_format = Column(String)
    auto_start = Column(Boolean, index=True, default=True)
    schema = relationship('Schema', back_populates='sessions')
    _connected: bool = False

    @staticmethod
    def clone(row):
        data = {}
        for column in row.__table__.columns:
            if column.name in ['id']:
                continue

            data[column.name] = getattr(row, column.name)

        return Session(**data)

    @staticmethod
    def delete(session: 'Session'):
        db_session.delete(session)
        db_session.commit()

    @staticmethod
    def find_by_id(session_id: int):
        cursor = db_session.query(Session)
        return cursor.filter(Session.id == session_id).first()

    @staticmethod
    def index(hostname=None):
        cursor = db_session.query(Session)
        if hostname:
            search = "%{}%".format(hostname)
            cursor = cursor.filter(Session.hostname.like(search))

        return cursor.all()

    @property
    def connected(self):
        return self._connected

    @connected.setter
    def connected(self, value):
        self._connected = value

    @property
    def tmux_id(self) -> str:
        fields = [
            'pfm_session',
            '{schema_name}',
            '{hostname}',
            '{remote_address}',
            '{remote_port}',
            '{local_address}',
            '{local_port}',
            '{type}'
        ]

        data = {
            'schema_name': self.schema.name,
            'hostname': self.hostname,
            'remote_address': self.remote_address,
            'remote_port': self.remote_port,
            'local_address': self.local_address,
            'local_port': self.local_port,
            'type': self.type
        }

        return '|'.join(fields).format(**data).replace('.', '_')

    @property
    def url(self):
        if self.url_format and self.connected:
            data = {
                'schema_id': self.schema_id,
                'hostname': self.hostname,
                'remote_address': self.remote_address,
                'remote_port': self.remote_port,
                'local_address': self.local_address,
                'local_port': self.local_port,
                'type': self.type
            }

            return self.url_format.format(**data)
        else:
            return ''

    @property
    def command(self):
        ssh_options = [
            '-o ExitOnForwardFailure=yes',
            '-o ServerAliveCountMax=3',
            '-o ServerAliveInterval=10'
        ]

        session_data = self.as_dict()

        session_data['tmux_id'] = self.tmux_id
        session_data['shell_command'] = 'ping localhost'
        session_data['options'] = ' '.join(ssh_options)

        if self.type == 'remote':
            ssh = 'ssh {options} -R {local_address}:{local_port}:{remote_address}:{remote_port} {hostname}'
        else:
            ssh = 'ssh {options} -L {local_address}:{local_port}:{remote_address}:{remote_port} {hostname}'

        session_data['ssh_command'] = ssh.format(**session_data)

        # start_command = "screen -dmS '{name}' {ssh_command}  -- {shell_command}"
        start_command = "tmux new-session -d -s '{tmux_id}' '{ssh_command} -- {shell_command}'".format(**session_data)
        # inspect(session_definition)
        return start_command

    @staticmethod
    def get_state():
        state = []
        for session in Session.index():
            session_state = session.as_dict()
            session_state['connected'] = session.connected
            session_state['url'] = session.url
            state.append(session_state)
        return state


def init_database():
    BaseModelDB.metadata.create_all(engine)
    group = Group.find_by_name('pfm-system')
    if not group:
        print(f"Group missing {group}")
        group = Group(name="pfm-system", label="System")
        schema = Schema(name="pfm-ephemeral", label="Ephemeral sessions")
        group.schemas.append(schema)
        Group.add(group)
        db_session.commit()


def reset_database():
    BaseModelDB.metadata.drop_all(engine)



import subprocess
import random
from rich import inspect, print
from rich.table import Table
from . import tools, models


def execute(command):
    # print(command)
    result = subprocess.run(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

    # inspect(result)
    # exit(1)
    if result.returncode != 0:
        print(f"Command [b]{command}[/] failed:\n {result.stderr}")
        exit(3)

    return result


def stop(session: models.Session):
    if session.connected:
        print("Stop session {tmux_id}.".format(tmux_id=session.tmux_id))
        stop_command = "tmux kill-session -t '{tmux_id}'".format(tmux_id=session.tmux_id)
        execute(stop_command)


def in_active_sessions(definition, active_sessions):
    for session in active_sessions:
        shared_items = {k: session[k] for k in session if k in definition and session[k] == definition[k]}
        # print(len(shared_items))
        if len(shared_items) >= 7:
            return True

    return False


def generate_new_port(port_list):
    while True:
        port = str(random.randint(9100, 9500))
        if port not in port_list:
            port_list.append(port)
            return port


def start(session, reconnect: bool = False):
    settings = tools.settings()
    if reconnect and session.connected:
        # print(f"[b]Stopping '{session_name}'[/b]")
        stop(session)

    if not session.connected:
        generate_port = False

        if not session.local_port:
            generate_port = True

        if session.local_port in settings.ports_active:
            if session.local_port_dynamic:
                print(f"[b]Warning[/b]: Local port {session.local_port} already in use, assigning new port")
                generate_port = True
            else:
                print(f"[b]Warning[/b]: Local port {session.local_port} already in use")
                exit()

        if generate_port:
            session.local_port = generate_new_port(settings.ports_active)

        result = execute(session.command)
        return result
    else:
        print("Ignoring {type} forward on [b]{hostname}[/b] {remote_address}:{remote_port}.".format(**session.as_dict()))


def refresh_state():
    update_state()
    for session in models.Session.index():
        if not session.active or session.connected:
            continue

        start(session)


def update_state():
    settings = tools.settings()

    # Reset state
    settings.ports_active = []
    for schema in models.Schema.index():
        # schema.active = False

        for session in schema.sessions:
            session.connected = False

    # Update state from tmux sessions
    command = "tmux ls | grep 'pfm_session' | cut -d ':' -f 1"
    result = execute(command)

    session_id_list = result.stdout.split("\n")

    for session_id in session_id_list:
        if len(session_id) == 0:
            continue

        session_data = session_id.replace('_', '.').replace(':', '')
        # print(f"Active -> {session_id}")
        values = session_data.split('|')

        if len(values) == 8:
            filler, schema_name, hostname, remote_host, remote_port, local_address, local_port, forward_type = values

            schema = models.Schema.find_by_name(schema_name)
            if schema:
                # schema.active = True
                session = schema.get_session(forward_type, hostname, int(remote_port))
                session.connected = True
                schema.active = True
                # session.local_address = local_address
                session.local_port = local_port
                settings.ports_active.append(local_port)
            else:
                print("TMUX session active but missing session definition.")


def filter_session(session: models.Session, schema_name: str = None, host: str = None, port: str = None):
    if schema_name and schema_name not in session.schema.name:
        return True
    if host and host not in session.hostname:
        return True
    if port and port != session.remote_port:
        return True


def list_from_active(key, filter_string: str = ''):
    items = []

    for schema in models.Schema.index():
        for session in schema.sessions:
            if not session.active:
                continue

            item = session.label
            if filter_string not in item:
                continue

            if item not in items:
                items.append(item)

    return items


def prepare_table(show_alias: bool = False, show_link: bool = False, show_schema: bool = True, title: str = 'Schemas'):
    settings = tools.settings()
    table = Table(show_edge=settings.show_table_border)

    if show_schema:
        table.add_column("Schema", style="green", width=30)

    table.add_column("Hostname", justify="left", style="yellow", width=30)
    table.add_column("Type", justify="right", style="yellow", width=8)
    table.add_column("Local address", justify="right", style="white", width=15)
    table.add_column("Local port", justify="right", style="white", width=7)
    table.add_column("Remote address", justify="right", style="blue", width=15)
    table.add_column("Remote port", justify="right", style="blue", width=7)

    if show_alias:
        table.add_column("Alias", justify="left", style="cyan", width=25)

    if show_link:
        table.add_column("URL", style="cyan", width=60)

    return table


def show_active_sessions(schema: str = None, host: str = None, port: int = None):
    update_state()

    table = prepare_table(True, True, True)
    session_count = 0
    for schema in models.Schema.index():
        schema_rows = False
        for session in schema.sessions:
            if not session.connected:
                continue

            schema_rows = True

            row = [
                f"{session.id:3d} {schema.name}",
                session.hostname,
                session.type,
                session.local_address,
                'auto' if session.local_port == 0 else session.local_port.__str__(),
                session.remote_address,
                session.remote_port.__str__(),
                session.label,
                session.url
            ]
            table.add_row(*row)
            session_count += 1
        if schema_rows:
            table.rows[table.row_count - 1].end_section = True

    if session_count > 0:
        print(table)
    else:
        print("Nothing to see here")




import os
from sqlalchemy import create_engine, Column, Integer
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.orm import sessionmaker

from . import tools

database_path = os.path.join(tools.base_path, 'database.db')
SQLALCHEMY_DATABASE_URL = f"sqlite:///{ database_path }"

engine = create_engine(SQLALCHEMY_DATABASE_URL, connect_args={"check_same_thread": False})


class BaseModel(object):
    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()

    def as_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}

    id = Column(Integer, primary_key=True)


SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)
BaseModelDB = declarative_base(cls=BaseModel)



import typer
import time
from enum import Enum
import rich
from rich.table import Table
from rich.prompt import Confirm

from . import database, models, tools
from .forward_sessions import update_state, start, stop

app = typer.Typer(no_args_is_help=True)
db = database.SessionLocal()


class SessionTypes(str, Enum):
    local = "local"
    remote = "remote"
    ssh = "ssh"


def prepare_sessions_table(show_link: bool = False):
    table = Table(show_edge=False)

    table.add_column("#", style="green", justify='right', width=3)
    table.add_column("Schema", style="green", width=30)
    table.add_column("Hostname", justify="left", style="yellow", width=30)
    table.add_column("Type", justify="right", style="yellow", width=8)
    table.add_column("Local address", justify="right", style="white", width=15)
    table.add_column("Local port", justify="right", style="white", width=7)
    table.add_column("Remote address", justify="right", style="blue", width=15)
    table.add_column("Remote port", justify="right", style="blue", width=7)
    table.add_column("Alias", justify="left", style="cyan", width=25)

    if show_link:
        table.add_column("URL", style="cyan", width=60)

    return table


def print_index(hostname: str = None, active_only: bool = False):
    table = prepare_sessions_table()
    for session in models.Session.index(hostname):
        if active_only and not session.connected:
            continue

        row = [
            str(session.id),
            session.schema.name,
            session.hostname,
            session.type,
            session.local_address,
            'auto' if session.local_port_dynamic else session.local_port.__str__(),
            session.remote_address,
            session.remote_port.__str__(),
            session.label
        ]

        table.add_row(*row)

    if table.row_count > 0:
        table.rows[table.row_count - 1].end_section = True
    rich.print(table)


def print_session(session: models.Session):
    table = prepare_sessions_table()
    row = [
        str(session.id),
        session.schema.name,
        session.hostname,
        session.type,
        session.local_address,
        'auto' if session.local_port_dynamic else session.local_port.__str__(),
        session.remote_address,
        session.remote_port.__str__(),
        session.label
    ]

    table.add_row(*row)

    table.rows[table.row_count - 1].end_section = True
    rich.print(table)


@app.command()
def index(hostname: str = typer.Argument(None, help='Criteria to search'),
          active: bool = typer.Option(False, help='Criteria to search')):
    """
    List sessions
    """
    update_state()
    print_index(hostname, active)


@app.command()
def create(schema_name: str = typer.Argument(None, help='Schema name'),
           session_type: SessionTypes = typer.Argument(SessionTypes.local, help='Session type', case_sensitive=False),
           label: str = typer.Argument(None, help='Session label'),
           hostname: str = typer.Argument(None, help='Session hostname'),
           remote_address: str = typer.Option('127.0.0.1', help='Remote address'),
           remote_port: int = typer.Option(0, help='Remote address', min=0, max=65535),
           local_address: str = typer.Option('127.0.0.1', help='Remote address'),
           local_port: int = typer.Option(0, help='Remote address', min=0, max=65535),
           local_port_dynamic: bool = typer.Option(True, help='Assign local port dynamically'),
           auto_start: bool = typer.Option(True, help='Remote address'),
           url_format: str = typer.Option('http://{hostname}:{local_port}', help='Dynamic URL'),
           schema_id: bool = typer.Option(True, help='Use schema ID instead of name')
           ):
    """
    Create session
    """
    if schema_id:
        schema = models.Schema.find_by_id(int(schema_name))
    else:
        schema = models.Schema.find_by_name(schema_name)
    if schema is None:
        print(f"Schema '{schema_name}' doesn't exist")
        exit(401)

    session = models.Session()
    session.type = session_type.value
    session.label = label
    session.hostname = hostname
    session.remote_address = remote_address
    session.remote_port = remote_port
    session.local_address = local_address
    session.local_port = local_port
    session.local_port_dynamic = local_port_dynamic
    session.auto_start = auto_start
    session.url_format = url_format
    schema.sessions.append(session)
    models.db_session.commit()
    print_session(session)


@app.command()
def delete(session_id: int = typer.Argument(None, help='Session ID'),
           force: bool = typer.Option(False, help="Force delete")):
    """Delete session"""
    session = models.Session.find_by_id(session_id)
    if force or Confirm.ask(f"Are you sure you want to delete '{session.label}'?"):
        models.Session.delete(session)


@app.command()
def update(session_id: int = typer.Argument(None, help='Session ID'),
           schema_name: str = typer.Option(None, help='Schema name'),
           session_type: SessionTypes = typer.Option(None, help='Session type', case_sensitive=False),
           label: str = typer.Option(None, help='Session label'),
           hostname: str = typer.Option(None, help='Session hostname'),
           remote_address: str = typer.Option(None, help='Remote address'),
           remote_port: int = typer.Option(None, help='Remote address', min=0, max=65535),
           local_address: str = typer.Option(None, help='Remote address'),
           local_port: int = typer.Option(None, help='Remote address', min=0, max=65535),
           local_port_dynamic: bool = typer.Option(None, help='Assign local port dynamically'),
           auto_start: bool = typer.Option(None, help='Remote address'),
           url_format: str = typer.Option(None, help='Dynamic URL'),
           schema_id: bool = typer.Option(True, help='Use schema ID instead of name')):
    """
    Update session settings
    """

    session = models.Session.find_by_id(session_id)
    if not session:
        print("Session ID not found")
        exit(401)

    if session_type:
        session.type = session_type.value

    if label:
        session.label = label

    if hostname:
        session.hostname = hostname

    if remote_address:
        session.remote_address = remote_address

    if remote_port:
        session.remote_port = remote_port

    if local_address:
        session.local_address = local_address

    if local_port:
        session.local_port = local_port

    if local_port_dynamic:
        session.local_port_dynamic = local_port_dynamic

    if auto_start:
        session.auto_start = auto_start

    if url_format:
        session.url_format = url_format

    if schema_name:
        if schema_id:
            schema = models.Schema.find_by_id(int(schema_name))
        else:
            schema = models.Schema.find_by_name(schema_name)
        if schema is None:
            print(f"Schema '{schema_name}' doesn't exist")
            exit(401)

        session.schema_id = schema.id

    models.db_session.commit()
    print_session(session)


@app.command()
def connect(label: str = typer.Argument(..., help='Session label'),
            hostname: str = typer.Argument(..., help='Session hostname'),
            remote_port: int = typer.Argument(..., help='Remote address', min=0, max=65535),
            session_type: SessionTypes = typer.Option(SessionTypes.local, help='Session type', case_sensitive=False),
            remote_address: str = typer.Option('127.0.0.1', help='Remote address'),
            local_address: str = typer.Option('127.0.0.1', help='Remote address'),
            local_port: int = typer.Option(0, help='Remote address', min=0, max=65535),
            ):
    """
    Connect ephemeral session
    """
    settings = tools.settings()

    schema_name = 'pfm-ephemeral'
    schema = models.Schema.find_by_name(schema_name)

    if schema is None:
        print(f"Schema '{schema_name}' doesn't exist")
        exit(401)

    session = models.Session()
    session.type = session_type.value
    session.label = label
    session.hostname = hostname
    session.remote_address = remote_address
    session.remote_port = remote_port
    session.local_address = local_address
    session.local_port = local_port
    session.local_port_dynamic = local_port == 0
    session.auto_start = False
    session.url_format = ''
    schema.sessions.append(session)
    models.db_session.commit()

    start(session, False)
    session.active = True
    session.connected = True

    time.sleep(settings.wait_after_start)
    models.db_session.commit()

    print_session(session)


@app.command()
def disconnect(session_id: int = typer.Argument(None, help='Session ID')):
    """
    Disconnect session
    """
    update_state()
    session = models.Session.find_by_id(session_id)

    if not session:
        print("Session doesn't exist")
        exit()

    if Confirm.ask(f"Are you sure you want to disconnect '{session.label}'?"):
        stop(session)

        if session.schema.name == 'pfm-ephemeral':
            print("Delete ADHOC session")
            models.Session.delete(session)



import os
from typing import List

import yaml
from pydantic import BaseModel
from rich import print

base_path = os.path.expanduser('~/.pfm')
if not os.path.isdir(base_path):
    os.mkdir(base_path)

settings_file_location = os.path.join(base_path, 'pfm3.settings.yaml')


class Settings(BaseModel):
    ports_active: List = []
    show_schema_link: bool = False
    show_table_border: bool = True
    wait_after_start: float = 0.5
    wait_after_stop: float = 0.5


_settings: Settings = Settings()


def settings():
    return _settings


def load_yaml_file(filename: str):
    with open(os.path.expanduser(filename), "r") as stream:
        try:
            return yaml.load(stream, Loader=yaml.FullLoader)
        except yaml.YAMLError as exc:
            print(exc)


def write_file(filename: str, data):
    with open(os.path.expanduser(filename), "w") as stream:
        try:
            stream.write(data)
        except yaml.YAMLError as exc:
            print(exc)


def load_settings():
    global _settings, settings_file_location
    # print("Load settings")

    if not os.path.isfile(settings_file_location):
        save_settings()

    loaded_settings = load_yaml_file(settings_file_location)
    # print(loaded_settings)
    _settings = Settings(**loaded_settings)
    # settings.update(loaded_settings)

    # if loaded_settings != settings:
    #    save_settings()


def save_settings():
    global _settings, settings_file_location

    print(f"[b]Updating configuration file on '{settings_file_location}'[/b]")
    with open(settings_file_location, "w") as stream:
        try:
            settings_data = _settings.dict()
            yaml.dump(settings_data, stream)
            print(settings_data)
        except yaml.YAMLError as exc:
            print(exc)


def set_ports_active(ports):
    _settings.ports_active = ports


def get_ports_active():
    return _settings.ports_active



from . import forward_sessions, models
from click.shell_completion import CompletionItem
from sshconf import read_ssh_config
from os.path import expanduser


def sc_schemas(ctx, param, incomplete):
    schemas = models.Schema.index()
    result = []

    for schema_name in schemas:
        if not schema_name.startswith(incomplete):
            continue

        schema = schemas.get(schema_name)
        item = CompletionItem(schema_name, help=f"{len(schema)} session(s) for {schema_name}")
        result.append(item)
    return result


def ac_schemas(incomplete):
    schemas = models.Schema.index()
    result = []

    for schema in schemas:
        if not schema.name.startswith(incomplete):
            continue

        item = (schema.name, f"{len(schema.sessions)} session(s) for {schema.name}")
        result.append(item)
    return result


def sc_active_schemas(ctx, param, incomplete):
    return forward_sessions.list_from_active('schema', incomplete)


def ac_active_schemas(incomplete):
    items = []
    for schema in models.Schema.index():
        if schema.active and schema.name not in items:
            items.append(schema.name)

    return items


def sc_active_remote_port(ctx, param, incomplete):
    return forward_sessions.list_from_active('remote_port')


def sc_active_hosts(ctx, param, incomplete):
    return forward_sessions.list_from_active('hostname')


def ac_hosts(incomplete):
    result = []
    c = read_ssh_config(expanduser("~/.ssh/config"))
    for host in c.hosts():
        if "*" not in host and host.startswith(incomplete):
            hosts = host.split(" ")
            for hostname in hosts:
                result.append(hostname)

    return result


def sc_hosts(ctx, param, incomplete):
    from sshconf import read_ssh_config
    from os.path import expanduser
    result = []
    c = read_ssh_config(expanduser("~/.ssh/config"))
    for host in c.hosts():
        if "*" not in host and host.startswith(incomplete):
            hosts = host.split(" ")
            for hostname in hosts:
                item = CompletionItem(hostname)
                result.append(item)

    return result



import os.path
import yaml
from sshconf import read_ssh_config
import pkg_resources
import simplejson
import typer
import time
from rich.prompt import Confirm
import rich

from .cli_autocomplete import sc_schemas, sc_hosts
from . import tools, forward_sessions, models, cli_group, cli_schema, cli_session, cli_ssh_group


app = typer.Typer(no_args_is_help=True)
app.add_typer(cli_group.app, name='group', help="Group management")
app.add_typer(cli_schema.app, name='schema', help="Schema management")
app.add_typer(cli_session.app, name='session', help="Session management")
app.add_typer(cli_ssh_group.app, name='ssh_group', help="SSH group management")
tools.load_settings()


@app.command()
def shutdown():
    """
    Stop all active sessions
    """

    settings = tools.settings()

    forward_sessions.update_state()
    for schema in models.Schema.index():
        for session in schema.sessions:
            if session.connected:
                forward_sessions.stop(session)
            session.active = False
        schema.active = False

    time.sleep(settings.wait_after_stop)
    forward_sessions.show_active_sessions()
    models.db_session.commit()


@app.command()
def status(schema: str = typer.Option(None, shell_complete=sc_schemas),
           host: str = typer.Option(None, shell_complete=sc_hosts),
           port: int = None,
           json: bool = typer.Option(False, '--json', '-j', help="Output JSON")):
    """
    Show active sessions
    """

    forward_sessions.show_active_sessions(schema, host, port)


@app.command()
def state():
    """
    Show current state in JSON format
    """

    ssh_config = read_ssh_config(os.path.expanduser("~/.ssh/config"))
    ssh_hosts = []
    for host in ssh_config.hosts():
        if "*" in host:
            continue
        hosts = host.split(" ")
        for hostname in hosts:
            ssh_hosts.append(hostname)

    forward_sessions.refresh_state()
    time.sleep(0.5)
    forward_sessions.update_state()
    current_state = {
        'groups': models.Group.get_state(),
        'schemas': models.Schema.get_state(),
        'sessions': models.Session.get_state(),
        'ssh_groups': models.SSHGroup.get_state(),
        'ssh_hosts': ssh_hosts
    }
    print(simplejson.dumps(current_state, indent=2))


@app.command()
def version():
    """
    Show PFM version
    """

    current_version = pkg_resources.get_distribution("port-forward-manager").version
    rich.print(f"Port Forward Manager [bold white]v{current_version}[/]")


@app.command()
def db_wipe():
    """Wipe the whole database clean"""
    if Confirm.ask("This action will wipe all groups, schemas and sessions, are you sure?"):
        models.reset_database()
        models.init_database()


@app.command()
def yaml_export(export_path: str = typer.Argument(None, help="YAML configuration file")):
    """Import groups, schemas and sessions from configuration file"""
    export_data = {
        'export_format': 'db_dump',
        'groups': models.Group.get_state(),
        'schemas': models.Schema.get_state(),
        'sessions': models.Session.get_state(),
        'ssh_groups': models.SSHGroup.get_state(),
    }

    yaml_string = yaml.dump(export_data)
    if export_path:
        tools.write_file(export_path, yaml_string)
    else:
        print(yaml_string)


@app.command()
def yaml_import(export_path: str = typer.Argument(..., help="YAML configuration file"),
                wipe: bool = typer.Option(False, help="Wipe DB")):
    """Import groups, schemas and sessions from configuration file"""
    settings = tools.load_yaml_file(export_path)

    change_count = 0

    if settings.get('export_format') != 'db_dump':
        rich.print("[red]Invalid export file format[/]")
        exit()

    if wipe:
        if Confirm.ask("This action will wipe all groups, schemas and sessions, are you sure?"):
            models.reset_database()
            models.init_database()

    rich.print("Importing PFM DB DUMP")
    groups = {}
    schemas = {}

    # Import groups
    current_groups = []
    for group_definition in settings.get('groups', []):
        group_name = group_definition.get('name')
        groups[group_definition.get('id')] = group_definition.get('name')
        group = models.Group.find_by_name(group_name)
        if not group:
            change_count += 1
            rich.print(f"* Creating group {group_name}")
            group = models.Group(name=group_name, label=group_definition.get('label'))
            models.Group.add(group)
        else:
            group.label = group_definition.get('label')
            rich.print(f"* Update group {group_name}")

    for group in models.Group.index():
        rich.print(f"* Checking group {group.name}")
        if group.id not in groups.keys():
            rich.print(f"  * Deleting group")
            models.Group.delete(group)

    rich.print(f"Imported {change_count} groups")
    models.db_session.commit()
    
    # Import schemas
    for schema_definition in settings.get('schemas', []):
        group_name = groups.get(schema_definition.get('group_id'))
        group = models.Group.find_by_name(group_name)
        if not group:
            rich.print(f"IGNORING schema - Error could not find group {group_name}")
            continue
        schemas[schema_definition.get('id')] = schema_definition.get('name')
        schema_name = schema_definition.get('name')

        schema = models.Schema.find_by_name(schema_name)
        if not schema:
            change_count += 1
            rich.print(f"* Importing schema {schema_name}")
            schema = models.Schema(name=schema_name,
                                   label=schema_definition.get('label'),
                                   active=False)

            group.schemas.append(schema)

    models.db_session.commit()

    # Import sessions
    for session_definition in settings.get('sessions'):
        schema_name = schemas.get(session_definition.get('schema_id'))
        schema = models.Schema.find_by_name(schema_name)
        if not schema:
            rich.print(f"IGNORING session - Error could not find schema '{schema_name}'")
            continue

        session_type = session_definition.get('type', 'local')
        hostname = session_definition.get('hostname')
        remote_port = session_definition.get('remote_port')
        if remote_port:
            remote_port = int(remote_port)

        session = schema.get_session(session_type, hostname, remote_port)
        if not session:
            change_count += 1
            print(f"    * Importing session {session_type} {hostname} {remote_port}")
            auto_start = session_definition.get('auto_start')
            if auto_start or auto_start == 'true':
                auto_start = True
            session = models.Session(label=session_definition.get('label'),
                                     type=session_type,
                                     active=False,
                                     hostname=session_definition.get('hostname'),
                                     remote_address=session_definition.get('remote_address', '127.0.0.1'),
                                     remote_port=session_definition.get('remote_port'),
                                     local_address=session_definition.get('local_address', '127.0.0.1'),
                                     local_port=session_definition.get('local_port'),
                                     local_port_dynamic=session_definition.get('local_port') is None,
                                     url_format=session_definition.get('url', ''),
                                     auto_start=auto_start)
            schema.sessions.append(session)

    # Import SSH groups
    for ssh_group in settings.get('sshGroups', []):
        schema_name = schemas.get(ssh_group.get('schema_id'))
        schema = models.Schema.find_by_name(schema_name)
        if not schema:
            rich.print(f"IGNORING session - Error could not find schema '{schema_name}'")
            continue

        group_name = ssh_group.get('group_name')
        label = ssh_group.get('label')

        group = schema.get_ssh_group(group_name)
        if not group:
            change_count += 1
            print(f"    * Importing SSH group {group_name} {label}")

            group = models.SSHGroup(label=label, group_name=group_name)
            schema.ssh_groups.append(group)

    models.db_session.commit()

    if change_count == 0:
        rich.print("No changes...")
    else:
        rich.print("There were {} items imported.".format(change_count))


def run():
    # models.reset_database()
    models.init_database()
    app()


if __name__ == "__main__":
    app()



import time

import rich
from rich.table import Table
from rich.prompt import Confirm
import typer
from sqlalchemy.exc import IntegrityError
from . import database, models, tools, forward_sessions
from .cli_autocomplete import ac_schemas, ac_active_schemas, sc_active_hosts, sc_active_remote_port
from .cli_session import prepare_sessions_table
from .forward_sessions import show_active_sessions
from sqlalchemy.orm.session import make_transient

app = typer.Typer(no_args_is_help=True)
db = database.SessionLocal()


def prepare_schema_table(show_alias: bool = False, show_link: bool = False, show_schema: bool = True, title: str = 'Schemas'):

    table = Table(show_edge=False)

    table.add_column("#", justify="right", style="green", width=3)
    table.add_column("Group", justify="left", style="yellow", width=20)
    table.add_column("Schema name", justify="left", style="yellow", width=30)
    table.add_column("Schema label", justify="right", style="yellow", width=30)

    return table


@app.command()
def create(group_id: int = typer.Argument(..., help='Group name'),
           schema_name: str = typer.Argument(..., help='Schema unique name'),
           label: str = typer.Argument(..., help='Schema label')):
    """
    Create a schema
    """
    group = models.Group.find_by_id(group_id)
    if group is None:
        print(f"Group '{group_id}' doesn't exist")
        exit(401)

    if models.Schema.find_by_name(schema_name):
        print(f"Schema '{schema_name}' already exists")
        exit(401)

    schema = models.Schema(name=schema_name, label=label, group_id=group.id)
    group.schemas.append(schema)

    models.db_session.commit()


@app.command()
def update(schema_id: int = typer.Argument(..., help='Schema ID'),
           group_id: int = typer.Option(None, help='Group ID'),
           name: str = typer.Option(None, help='Schema unique name'),
           label: str = typer.Option(None, help='Schema label')):
    """
    Create a schema
    """
    schema = models.Schema.find_by_id(schema_id)
    if schema is None:
        print(f"Schema '{schema_id}' doesn't exist")
        exit(401)

    if group_id:
        group = models.Group.find_by_id(group_id)
        if not group:
            print("Group is invalid")
        else:
            schema.group_id = group_id

    if name:
        schema.name = name

    if label:
        schema.label = label

    models.db_session.commit()


@app.command()
def index(schema_filter: str = typer.Argument(None, autocompletion=ac_schemas)):
    """
    List configured schemas
    """
    settings = tools.settings()
    table = prepare_schema_table(True, settings.show_schema_link, True)

    for schema in models.Schema.index(schema_filter):
        if schema.group_id is None:
            print(f"Schema #{schema.id} group is INVALID")
            row = [
                str(schema.id),
                'INVALID GROUP',
                schema.name,
                schema.label,
            ]
        else:
            row = [
                str(schema.id),
                schema.group.name,
                schema.name,
                schema.label,
            ]
        table.add_row(*row)

    if table.row_count > 0:
        table.rows[table.row_count - 1].end_section = True

    rich.print(table)


@app.command()
def clone(schema_id: int = typer.Argument(..., help='Schema ID'),
          group_id: int = typer.Option(None, help='Group ID'),
          name: str = typer.Option(None, help='Schema name'),
          label: str = typer.Option(None, help='Schema label'),):

    schema_original = models.Schema.find_by_id(schema_id)
    if schema_original is None:
        print(f"Schema {schema_id} not found")
        exit(3)

    if group_id is None:
        group_id = schema_original.group_id

    group = models.Group.find_by_id(group_id)
    if not group:
        print(f"Group {group_id} not found")
        exit(3)

    print(f"Clone schema {schema_original.name}")
    if not name:
        name = f"{schema_original.name}_clone"

    if not label:
        label = f"{schema_original.label} clone"
    schema_clone = models.Schema(name=name, label=label)
    group.schemas.append(schema_clone)

    for session_original in schema_original.sessions:
        session_clone = models.Session.clone(session_original)
        if session_clone.local_port_dynamic:
            session_clone.local_port = 0

        schema_clone.sessions.append(session_clone)

    models.db_session.commit()


@app.command()
def delete(schema_id: int = typer.Argument(None, help='Schema ID'),
           force: bool = typer.Option(False, help="Force delete")):
    """Delete session"""

    schema = models.Schema.find_by_id(schema_id)
    if schema is None:
        print("Schema not found")
        exit()

    if force or Confirm.ask(f"Are you sure you want to delete '{schema.label}'?"):
        models.Schema.delete(schema)


@app.command()
def sessions(schema_filter: str = typer.Argument(None, autocompletion=ac_schemas)):
    """
    List configured schemas
    """
    settings = tools.settings()
    table = prepare_sessions_table()

    for schema in models.Schema.index(schema_filter):
        for session in schema.sessions:
            row = [
                str(schema.id),
                schema.name,
                session.hostname,
                session.type,
                session.local_address,
                '-----' if session.local_port_dynamic else session.local_port.__str__(),
                session.remote_address,
                session.remote_port.__str__(),
                session.label
            ]
            table.add_row(*row)
        if table.row_count > 0:
            table.rows[table.row_count - 1].end_section = True

    rich.print(table)


@app.command()
def start(schema_name: str = typer.Argument(..., autocompletion=ac_schemas),
          force: bool = typer.Option(None, help="Force sessions reconnection")):
    """
    Start a schema of forwarding sessions
    """
    settings = tools.settings()
    forward_sessions.update_state()

    schema = models.Schema.find_by_name(schema_name)

    if schema is None:
        print("[b]Schema '{0}' is unknown[/b]".format(schema_name))
        exit(-1)

    schema.active = True

    for session in schema.sessions:
        session.schema_id = schema.id
        forward_sessions.start(session, force)
        session.active = True

    time.sleep(settings.wait_after_start)
    show_active_sessions()
    models.db_session.commit()


@app.command()
def stop(schema_name: str = typer.Argument(None, autocompletion=ac_active_schemas),
         hostname: str = typer.Option(None, shell_complete=sc_active_hosts),
         port: str = typer.Option(None, shell_complete=sc_active_remote_port)):
    """
    Stop sessions from active schema, host or port
    """

    if not schema_name and not hostname and not port:
        print("[b]Pick a schema, host or host and port or --all[/b]")
        exit(-1)

    settings = tools.settings()
    forward_sessions.update_state()

    if schema_name:
        schema = models.Schema.find_by_name(schema_name)
        for session in schema.sessions:

            forward_sessions.stop(session)
            session.active = False
        schema.active = False

    time.sleep(settings.wait_after_stop)
    show_active_sessions()
    models.db_session.commit()
